/****************************************************************************
 Sort a marked block, calling either the internal (slow) sort on small
 blocks, and the external (fast) sort on large blocks.

 08 May 2019 SEM - check MsgLevel before outputting "Finished" message.

 04 Nov 2011 SEM - bump sort threshold to 6000 (machines are faster now).

 06 Dec 2010 SEM - bump sort threshold to 4000 (machines are faster now).

 28 Feb 2009 SEM - Fix problem of not finding tsort.com in certain cases where
           TSELOADDIR was used. Reported (along with a suggested fix) by Ross
           Barnes.

 Oct 24 2006 SEM: better error message on failure.

 2-04-2000 SEM: tsort.com is an old dos-based program.  Maximum command line
           that it supports is 128 characters.  This can be a problem, since
           tsort requires an input and output filename.  A solution is to
           switch to the drive/directory where the sorting will occur, and
           then use relative filenames.

11-16-99   SEM: Use either temp drive or drive file to be sorted is on as
           sort work drive, depending on which one has the most free disk
           space.

 2-10-99   SEM: Replaced OS checking with new built-in command. Made error
           messages a little prettier.

 8-21-98   SEM: Set the sorts' temporary file drive to the same as the drive
           where the editor writes the file to be sorted.  This is the
           SwapPath editor variable for the DOS version, and the TEMP
           environment variable path for the Win32 version.

 8-17-98   SEM: Bug!  The ignore-case option was being set the opposite of
           what it should be.

 3-27-98   SEM: Quote the path of the sort executable, if needed.

 2-05-98   SEM: Add "kill duplicates" option, and accept two different
           command line formats.  New format is:

           sort -i[gnorecase] -d[escending] -k[illdups]

                    or

           sort integer_sort_flags
                where flags can be _DESCENDING_ and/or _IGNORE_CASE_

           If no options passed, then sort will be ascending, respecting case.

 1-08-98   SEM: Tsort fails when the temp path is a long filename. Since
           tsort is a 16 program that does not handle long filenames, use
           GetShortPathName() to convert long names to short names before
           passing them to tsort.

 9-39-97   SEM: check to see if we are Windows NT or Windows 95.  If NT,
           call the sort via Dos(), if 95 call the sort via lDos(). Note
           that the options for the Dos() command to keep the window from
           showing are very particular, it took quite a bit of testing to
           come up with the right combination.

 7-30-97   SEM: win32 CreateProcess() doesn't always work with 16-bit
           executables.  In lieu of that, go back to calling Dos.  However,
           using the proper options, the screen doesn't get resized anymore!

 1-09-96   SEM: Always erase the block from disk (even if sort fails)

12-13-96   SEM: Use lDos in lieu of Dos to keep screen from resizing

12-18-95   SEM: Reset blocktype even if an error occurs.

           Call internal (slow) sort if:

           -lines to sort is < sort_thresh
           -file is in binary mode
           -block contains tabs or cr/lf characters
           -external sort module can not be found

04-27-20   SEM: Remove explicit Win95/98 support

09-30-2022 SEM: using CompareLines() from the ignore-case on kill dupes.
            Pass the flags in, and respect ignore case.

 ****************************************************************************/

// return codes:

constant
    ERROR_SAVEBLOCK = 0ff00h,   // SaveBlock fails
    ERROR_SPAWN     = 0fe00h,   // if the Dos/lDos command fails
    ERROR_GONE      = 0fc00h,   // everything looks ok, but sorted file does not exist
    ERROR_DISK_FULL = 00036h    // out of disk space

#define MAXPATH 255

dll "<kernel32.dll>"
    integer proc GetLastError() : "GetLastError"
    integer proc GetDiskFreeSpace(string root_dir:cstrval,
                    var integer SectorsPerCluster,
                    var integer BytesPerSector,
                    var integer NumberOfFreeClusters,
                    var integer TotalNumberOfClusters) : "GetDiskFreeSpaceA"
end

integer proc mGetDiskFreeSpace(string p)
    integer sectors_per_cluster, bytes_per_sector, free_clusters, total_clusters

    if GetDiskFreeSpace(p[1:3], sectors_per_cluster, bytes_per_sector, free_clusters, total_clusters)
        return (bytes_per_sector * sectors_per_cluster * free_clusters)
    endif
    return (0)
end

string proc MostFreeSpace(string p1, string p2)
    integer
        f1 = mGetDiskFreeSpace(p1),
        f2 = mGetDiskFreeSpace(p2)

    //  both >= 0
    if f1 >= 0 and f2 >= 0
        return (iif(f1 >= f2, p1, p2))
    endif

    //  both < 0,
    if f1 < 0 and f2 < 0
        return (iif(f1 <= f2, p1, p2))
    endif

    // 1 is < 0
    return (iif(f1 < 0, p1, p2))
end

proc mWarn(string msg)
    MsgBox("Error!", msg)
end

proc mMessage(string msg)
    if Query(MsgLevel) == _ALL_MESSAGES_
        Message(msg)
    endif
end

integer proc remove_dups(integer flags, integer doit, integer line1, integer line2)
    integer n = 0, deleted = 0, ignore_case = 0

    if not doit or line1 >= line2
        return (0)
    endif

    PushPosition()

    GotoLine(line1)
    if flags & _IGNORE_CASE_
        ignore_case = 1
    endif

    mMessage("Removing duplicates")
    while CurrLine() < line2 and CurrLine() < NumLines()
        if CompareLines(CurrLine(), CurrLine() + 1, ignore_case) == 0
            Down()
            deleted = deleted + 1
            if not KillLine()
                break
            endif
            Up()
        elseif not Down()
            break
        endif
        n = n + 1
        if n mod 100000 == 0
            KeyPressed()
            Message("Removing dupes: ", CurrLine(), ' of ', NumLines())
        endif
    endwhile

    PopPosition()
    return (deleted)
end

/**************************************************************************
  Decode the passed options - can either be a binary integer, or a flag in
  the format -option

  Return TRUE if the -killdups option is present
 **************************************************************************/
integer proc GetOptions(string cmdline, var integer flags, var string options)
    integer i, n, z, kill_dups, descending, ignore_case
    string s[32]

    kill_dups = FALSE
    descending = FALSE
    ignore_case = FALSE
    n = NumTokens(cmdline, ' ')
    for i = 1 to n
        s = GetToken(cmdline, ' ', i)
        case Lower(s[1:2])
            when '-i'
                ignore_case = TRUE
            when '-d'
                descending = TRUE
            when '-k'
                kill_dups = TRUE
            otherwise
                z = Val(s)
                if z & _IGNORE_CASE_
                    ignore_case = TRUE
                endif
                if z & _DESCENDING_
                    descending = TRUE
                endif
        endcase
    endfor

    if ignore_case
        flags = flags | _IGNORE_CASE_
    else
        options = options + "/A "
    endif

    if descending
        options = options + "/R "
        flags = flags | _DESCENDING_
    endif

    if kill_dups
        options = options + "/D "
    endif

    return (kill_dups)
end

proc mGotoLineCol(integer line, integer col)
    GotoLine(line)
    GotoColumn(col)
end

integer proc ChangeDriveDir(string dir)
    return (LogDrive(dir[1]) and Upper(GetDrive()) == Upper(dir[1]) and ChDir(dir))
end

proc mMarkColumn(integer y1, integer x1, integer y2, integer x2)
    UnmarkBlock()
    mGotoLineCol(y1, x1)
    MarkColumn()
    mGotoLineCol(y2, x2)
    MarkColumn()
end

/**************************************************************************
  Call the internal (slow) sort.

  Do not call it if the kill_dups option is set.
 **************************************************************************/
proc BuiltInSort(integer flags, integer kill_dups, string msg, integer y1, integer y2)
    integer result, deleted

    if msg == ""
        result = 1
    else
        result = MsgBox("", msg + " Use slow sort?", _YES_NO_CANCEL_)
    endif

    if result == 1
        Sort(flags)
        deleted = remove_dups(flags, kill_dups, y1, y2)
        if deleted > 0
            mMessage(Format(deleted; "duplicates removed"))
        else
            mMessage("Finished")
        endif
    else
        Message("Sort terminated")
    endif
end

integer sort_thresh = 6000
string sort_path[MAXPATH], sort_name[] = "tsort.com"

string fn[MAXPATH], efn[MAXPATH], options[20], sort_cmdline[255]
string dir[MAXPATH], save_dir[MAXPATH], t_dir[MAXPATH], drive[1]

proc main()
    integer x1, y1, x2, y2, b, line, col, row, marking, save_eoltype,
        flags, save_eoftype, ctl_char_found, kill_dups, rc, err_code
    integer error_buf, errors_found, deleted

    options = ""
    flags = 0
    errors_found = false

    kill_dups = GetOptions(Trim(Query(MacroCmdLine)), flags, options)
    b = isBlockInCurrFile()
    y1 = Query(BlockBegLine)
    y2 = Query(BlockEndLine)
    x1 = Query(BlockBegCol)
    x2 = Query(BlockEndCol)
    // no block, or just a small block
    if b == 0 or y2 - y1 + 1 < sort_thresh
        BuiltInSort(flags, kill_dups, "", y1, y2)
        return ()
    endif

    // can't find the external sort!
    if Length(sort_path) == 0
        sort_thresh = MAXINT
        BuiltInSort(flags, kill_dups, sort_name + " not found.", y1, y2)
        return ()
    endif

    // external sort doesn't handle binary files
    if BinaryMode()
        BuiltInSort(flags, kill_dups, "File is binary.", y1, y2)
        return ()
    endif

    // external sort would have problems with embedded control chars
    PushPosition()
    Message("Searching for Ctrl characters")
    ctl_char_found = lFind("[\n\r\t]", "glx")
    PopPosition()
    if ctl_char_found
        BuiltInSort(flags, kill_dups, "Ctrl characters found.", y1, y2)
        return ()
    endif

    // force eoltype to cr and lf, and force a cr and lf at eof
    save_eoltype = Set(EOLType, 3)
    save_eoftype = Set(EOFType, 2)

    marking = Query(Marking)
    line = CurrLine()
    col = CurrCol()
    row = CurrRow()

    if b <> _LINE_
        if b == _COLUMN_
            options = Format(options, "/+", x1, ":", x2 - x1 + 1)
        endif
        MarkLine(y1, y2)
    endif
    GotoBlockBegin()
    BegLine()

    dir = GetEnvStr("TEMP")
    if Length(dir) == 0
        dir = GetEnvStr("TMP")
        if Length(dir) == 0
            dir = CurrDir()
        endif
    endif

    dir = SplitPath(ExpandPath(dir), _DRIVE_|_PATH_)
    t_dir = dir
    dir = MostFreeSpace(t_dir, SplitPath(CurrFilename(), _DRIVE_|_PATH_))
    drive = dir[1]

    save_dir = CurrDir()
    if (Upper(dir[1]) in 'A'..'Z') and ChangeDriveDir(dir)
        dir = ".\"
    else
        save_dir = ""
    endif

    fn  = SplitPath(MakeTempName(dir), _NAME_|_EXT_)
    efn = SplitPath(MakeTempName(dir), _NAME_|_EXT_)

    fn = GetShortPath(fn)
    efn = GetShortPath(efn)

    Message("Writing out file to sort...")
    /**************************************************************************
      Save the block to be sorted in the temp directory
     **************************************************************************/
    rc = ERROR_SAVEBLOCK
    err_code = 0
    if SaveBlock(fn, _OVERWRITE_)
        rc = ERROR_SPAWN

        // get the free space again, after the file has been saved
        dir = MostFreeSpace(t_dir, SplitPath(CurrFilename(), _DRIVE_|_PATH_))
        drive = dir[1]

        sort_cmdline = Format("/Q /E", efn; "/T", drive; "/Z"; fn; fn; options)

        Message("Calling external sort...")
        if lDos(sort_path, sort_cmdline, _DONT_CLEAR_|_RETURN_CODE_|_START_HIDDEN_|_RUN_DETACHED_)
            rc = DosIOResult()
        else
            err_code = GetLastError()
        endif

        if rc == 0
            if not FileExists(fn)
                rc = ERROR_GONE
            else
                KillBlock()
                InsertFile(fn)
            endif
        endif

        PushPosition()
        error_buf = EditBuffer(efn)
        if NumLines() > 1 or CurrLineLen() > 0
            errors_found = true
        endif
        PopPosition()
        EraseDiskFile(efn)
        EraseDiskFile(fn)

        if save_dir <> ""
            ChangeDriveDir(save_dir)
        endif
    endif

    if b <> _COLUMN_
        MarkLine(y1, y2)
    elseif not marking
        MarkColumn(y1, x1, y2, x2)
    elseif line == y1 and col == x2
        mMarkColumn(y2, x1, y1, x2)
    elseif line == y2 and col == x1
        mMarkColumn(y1, x2, y2, x1)
    else
        MarkColumn(y1, x1, y2, x2)
    endif

    GotoLine(line)
    GotoColumn(col)
    ScrollToRow(row)
    Set(Marking, marking)

    case rc
        when ERROR_SAVEBLOCK
            mWarn(Format("Error on SaveBlock(", fn, ")"))
        when ERROR_SPAWN
            mWarn(Format("Error ", err_code, " running ", sort_path))
            mWarn(Format("Command line:"; sort_cmdline))
        when ERROR_GONE
            mWarn(Format("Error, sorted file disappeared!"))
        when ERROR_DISK_FULL
            mWarn(Format("Sort ran out of disk space on drive ", drive, ":"))
        when 0
            // no error
        otherwise
            mWarn(Format("Unknown Error"; rc:4:'0':16; "running external sort program:", sort_path))
            mWarn(Format("Cmdline:", sort_cmdline))
    endcase

    if errors_found
        MsgBoxBuff("Sort error", _OK_, error_buf)
    elseif rc == 0
        deleted = remove_dups(flags, kill_dups, y1, y2)
        if deleted > 0
            mMessage(Format(deleted; "duplicates removed"))
        else
            mMessage("Finished")
        endif
    endif
    AbandonFile(error_buf)

    Set(EOLType, save_eoltype)
    Set(EOFType, save_eoftype)
    PurgeMacro(CurrMacroFilename())
end

/**************************************************************************
  See if the specified pgm exists in the actual (LoadDir(1)) or the
  (possibly redirected) LoadDir().
 **************************************************************************/
proc find_editor_program(string fn, var string path)
    path = SplitPath(LoadDir(1), _DRIVE_|_PATH_) + fn
    if not FileExists(path)
        path = LoadDir() + fn
        if not FileExists(path)
            path = ""
        endif
    endif
    path = QuotePath(path)
end

proc WhenLoaded()
    find_editor_program("tsort.com", sort_path)
end
